(* Author:  Robbert van Renesse, Cornell University.
 *
 * See the file README to see what this is supposed to do.  This code is a mess.
 * When I started developing this, I didn't know what I was doing.  Now, if I
 * had to do it again, it could be a lot simpler...  I'll work on the commenting
 * this code better sometime, although I really should rewrite the whole thing.
 *)

open Printf
open Camouglob

(* This structure maintains some stuff that a lot of the routines
 * need.
 *)
type env = {
  dota:		out_channel;	(* the _post file (a for after) *)
  dotb:		out_channel;	(* the _pre file (b for before) *)
  dotc:		out_channel;	(* the .c file *)
  doth:		out_channel;	(* the .h file *)
  modname:	string		(* the name of the module *)
}

(* This counter is used to create unique names.
 *)
let uid_counter = ref 0

(* All the names that are locally defined are listed here.
 *)
let local_def = ref []

(* When we come across a new module name, we have to #include the
 * corresponding .h file.  Here we keep a list of the ones that
 * we've already included.
 *)
let included = ref []

(* We keep a cache of types so we can reuse them as much as
 * possible, reducing produced code.
 *)
let type_cache = ref []

(* We test a lot for unit types, which have to be treated
 * specially.
 *)
let unit_type = NAMED_TYPE(("", "unit"), [])

(* Capitalize the first character of s and return the result.
 *)
let capitalize s =
  let c = String.copy s in
  let first = String.get c 0 in
  begin
    match first with
      | 'a'..'z' ->
          String.set c 0 (Char.chr ((Char.code first) - (Char.code 'a') + (Char.code 'A')))
      | _ ->
	  ()
  end;
  c

(* Uncapitalize the first character of s and return the result.
 *)
let uncapitalize s =
  let c = String.copy s in
  let first = String.get c 0 in
  begin
    match first with
      | 'A'..'Z' ->
          String.set c 0 (Char.chr ((Char.code first) - (Char.code 'A') + (Char.code 'a')))
      | _ ->
	  ()
  end;
  c

let rec type_cmp =
  let cmp_lists tl1 tl2 =
    try
      let tl = List.combine tl1 tl2 in
      List.for_all type_cmp tl
    with _ ->
      false
  in function
  | (NAMED_TYPE((m1, n1), tl1)), (NAMED_TYPE((m2, n2), tl2)) ->
      if m1 <> m2 or n1 <> n2 then
	false
      else
	cmp_lists tl1 tl2
  | (ALPHA_TYPE _), (ALPHA_TYPE _) ->
      true
  | (FUNCTION_TYPE(args1, result1)), (FUNCTION_TYPE(args2, result2)) ->
      (cmp_lists args1 args2) && (type_cmp (result1, result2))
  | (PRODUCT_TYPE pl1), (PRODUCT_TYPE pl2) ->
      cmp_lists pl1 pl2
  | (STRUCT_TYPE sl1), (STRUCT_TYPE sl2) ->
      failwith "not implemented cmp struct yet"
  | (CONSTR_TYPE cl1), (CONSTR_TYPE cl2) ->
      failwith "not implemented cmp constr yet"
  | _ ->
      false

(* This function prints the variable name n in ml format.
 * If m = "", this means that no module was specified.
 * If it's locally defined in that case, prepend the
 * module name.  Otherwise it is probably a ``pervasive''
 * name (such as "int"), and no module name should be
 * prepended.  If the module name m is explicitly
 * specified, prepend the module name.
 *)
let print_ml_name oc modname m n =
  if m = "" then
    if List.mem n !local_def then
      fprintf oc "%s.%s" modname n
    else
      fprintf oc "%s" n
  else
    fprintf oc "%s.%s" m n

(* Same as print_ml_name, but for producing a C value.
 * In this case we use '_' to simulate modules.
 *)
let print_c_name oc modname m n =
  if m = "" then
    if List.mem n !local_def then
      if n = "t" then
        fprintf oc "%s_t" modname
      else
        fprintf oc "%s_%s_t" modname n
    else
      fprintf oc "%s" n
  else
    if n = "t" then
      fprintf oc "%s_t" m
    else
      fprintf oc "%s_%s_t" m n

(* This function transforms a list into a list of pairs,
 * where the lefthand side is an index, and the righthand
 * side is the corresponding entry in the original list.
 *)
let rec enumerate i = function
  | hd :: tl ->
      (i, hd) :: (enumerate (i + 1) tl)
  | [] ->
      []

(* Given a list of types, produce a list of tuples with
 * three entries:  the first is an index, the second is
 * a unique name for the entry in a structure called name,
 * and the third is the corresponding type.
 *)
let rec make_prod_list i name = function
  | hd :: tl ->
      (i, sprintf "%s.e%d" name i, hd) ::
	(make_prod_list (i + 1) name tl)
  | [] ->
      []

(* Like make_prod_list, but create argument names aj, a(j + 1), ...
 *)
let rec make_arg_list i j = function
  | hd :: tl ->
      (i, sprintf "a%d" j, hd) :: (make_arg_list (i + 1) (j + 1) tl)
  | [] ->
      []

(* Given a list of (name, _, type), produce a list like in
 * make_prod_list or make_arg_list using those names and types.
 *)
let rec make_struct_list i name = function
  | (n, _, t) :: tl ->
      (i, sprintf "%s.%s" name n, t) ::
	(make_struct_list (i + 1) name tl)
  | [] ->
      []

(* This function takes a print function, a separator string,
 * and a list of things to be printed.  The separator is
 * places between any two items.
 *)
let rec print_list oc pt sep = function
  | [] ->
      ()
  | [t] ->
      pt t
  | hd :: tl ->
      pt hd;
      fprintf oc "%s" sep;
      print_list oc pt sep tl

(* Print a C declaration for the given name and type.  Use
 * the given indentation indent.  If ind is true, make functions
 * indirect.  Otherwise leave them be.
 *)
let rec print_c_decl oc modname name indent ind = function
  | FUNCTION_TYPE(args, result) ->
      let print_arg (_, n, t) =
	print_c_decl oc modname n (indent ^ "    ") true t
      in
      let name = if ind then sprintf "(*%s)" name else name in
      if result = unit_type then
        fprintf oc "%svoid %s" indent name
      else
        print_c_decl oc modname name indent true result;
      begin
        match args with
	  | [NAMED_TYPE(("", "unit"), [])] ->
	      fprintf oc "(void)"
	  | _ ->
	      fprintf oc "(\n";
              print_list oc print_arg ",\n" (make_arg_list 0 0 args);
	      fprintf oc "\n%s)" indent
      end
  | NAMED_TYPE(("", "unit"), []) ->
      fprintf oc "%sunit_t %s" indent name
  | NAMED_TYPE(("", "float"), []) ->
      fprintf oc "%sdouble %s" indent name
  | NAMED_TYPE((m, n), []) ->
      fprintf oc "%s" indent;
      print_c_name oc modname m n;
      fprintf oc " %s" name
  | NAMED_TYPE(("", "ref"), [t]) ->
      print_c_decl oc modname (sprintf "*(%s)" name) indent true t
  | NAMED_TYPE(("", "array"), [t]) ->
      fprintf oc "%sstruct %s {\n" indent name;
      fprintf oc "%s    int size;\n" indent;
      print_c_decl oc modname "*ent" (indent ^ "    ") true t;
      fprintf oc ";\n%s} *%s" indent name
  | NAMED_TYPE(("", "list"), [t]) ->
      fprintf oc "%sstruct %s {\n" indent name;
      fprintf oc "%s    struct %s *next;\n" indent name;
      print_c_decl oc modname "ent" (indent ^ "    ") true t;
      fprintf oc ";\n%s} *%s" indent name
  | ALPHA_TYPE _ ->
      fprintf oc "%salpha %s" indent name
  | STRUCT_TYPE tl ->
      let print_struct (n, m, t) =
	print_c_decl oc modname n (indent ^ "    ") true t;
	fprintf oc ";\n"
      in
      fprintf oc "%sstruct %s {\n" indent name;
      List.iter print_struct tl;
      fprintf oc "%s} %s" indent name
  | PRODUCT_TYPE tl ->
      let print_factor (i, t) =
        print_c_decl oc modname (sprintf "e%d" i) (indent ^ "    ") true t;
	fprintf oc ";\n"
      in
      fprintf oc "%sstruct {\n" indent;
      List.iter print_factor (enumerate 0 tl);
      fprintf oc "%s} %s" indent name
  | CONSTR_TYPE cl ->
      let print_enum = function
	| CCONST n -> fprintf oc "%s_%s" modname n
	| NCCONST(n, _) -> fprintf oc "%s_%s" modname n
      and print_union = function
	| CCONST n -> ()
	| NCCONST(n, t) ->
	    print_c_decl oc modname n (indent ^ "        ") true t;
	    fprintf oc ";\n"
      in
      let ncconst = function
	| CCONST _ -> false
	| NCCONST(_, _) -> true
      in
      fprintf oc "%sstruct {\n%s    enum {" indent indent;
      print_list oc print_enum ", " cl;
      fprintf oc "} tag;\n";
      if List.exists ncconst cl then
	begin
	  fprintf oc "%s    union {\n" indent;
	  List.iter print_union cl;
	  fprintf oc "%s    } u;\n" indent
	end;
      fprintf oc "%s} %s" indent name
  | ABSTRACT_TYPE ->
      fprintf oc "%sabstract_t %s" indent name
  | _ ->
      fprintf oc "%s??? %s" indent name

(* Print an ML type for the given type.
 *)
let rec print_ml_type oc modname extern = function
  | ALPHA_TYPE n ->
      fprintf oc "'%s" n
  | FUNCTION_TYPE(args, result) ->
      if extern then
        fprintf oc "mlc"
      else
        let print_arg t =
	  fprintf oc "(";
          print_ml_type oc modname false t;
	  fprintf oc ") -> "
        in
        begin
          fprintf oc "(";
          List.iter print_arg args;
          fprintf oc "(";
          print_ml_type oc modname false result;
          fprintf oc "))"
	end
  | PRODUCT_TYPE tl ->
      fprintf oc "(";
      print_list oc (print_ml_type oc modname extern) " * " tl;
      fprintf oc ")"
  | NAMED_TYPE((m, n), tl) ->
      if tl <> [] then
	begin
	  fprintf oc "(";
	  print_list oc (print_ml_type oc modname extern) ", " tl;
	  fprintf oc ") "
	end;
      print_ml_name oc modname m n;
  | CONSTR_TYPE cl ->
      let print_constr = function
        | CCONST n ->
	    fprintf oc "%s" n
        | NCCONST(n, t) ->
	    fprintf oc "%s of " n;
	    print_ml_type oc modname extern t
      in
      print_list oc print_constr " | " cl
  | STRUCT_TYPE sl ->
      let print_struct (n, m, t) =
        if m then
          fprintf oc "mutable ";
        fprintf oc "%s: " n;
        print_ml_type oc modname extern t
      in
      fprintf oc "{";
      print_list oc print_struct "; " sl;
      fprintf oc "}"
  | ABSTRACT_TYPE ->
      ()

(* This function takes a name and a type, a produces C
 * code to convert from the ml type to the C type for
 * the variable of that name and type.  dir is an
 * argument to the conversion function that specifies
 * whether we're converting to an argument or a result.
 *)
let convert_ml_to_c oc modname name dir = function
  | NAMED_TYPE((m, n), []) ->
      fprintf oc "ml2c_";
      print_c_name oc modname m n;
      fprintf oc "(%s, %s)" dir name
  | ALPHA_TYPE _ ->
      fprintf oc "ml2c_alpha(%s)" name
  | ABSTRACT_TYPE ->
      fprintf oc "ml2c_abstract(%s)" name
  | FUNCTION_TYPE(_, _) ->
      fprintf oc "ml2c_upcall(%s, %s)" dir name
  | _ as t ->
      fprintf oc "UNKNOWN1(name = %s, dir = %s, t = " name dir;
      print_ml_type oc modname false t;
      fprintf oc ")"

let ml_to_c oc modname name dir = function
  | NAMED_TYPE(("", "list"), [t]) ->
      fprintf oc "    if (v == Val_int(0))\n";
      fprintf oc "        result = 0;\n";
      fprintf oc "    else {\n";
      fprintf oc "        result = (%s) malloc(sizeof(*result));\n" name;
      fprintf oc "        result->ent = ";
      convert_ml_to_c oc modname "Field(v, 0)" dir t;
      fprintf oc ";\n";
      fprintf oc "        result->next = ml2c_%s(%s, Field(v, 1));\n" name dir;
      fprintf oc "    }\n"
  | NAMED_TYPE(("", "array"), [t]) ->
      fprintf oc "    int i;\n\n";
      fprintf oc "    result = (%s) malloc(sizeof(*result));\n" name;
      fprintf oc "    result->size = Wosize_val(v);\n";
      fprintf oc "    result->ent = malloc(result->size * sizeof(*result->ent));\n";
      fprintf oc "    for (i = 0; i < result->size; i++)\n";
      fprintf oc "        result->ent[i] = ";
      convert_ml_to_c oc modname "Field(v, i)" dir t;
      fprintf oc ";\n"
  | PRODUCT_TYPE pl ->
      let print (i, n, t) =
	fprintf oc "    %s = " n;
	convert_ml_to_c oc modname (sprintf "Field(v, %d)" i) dir t;
	fprintf oc ";\n"
      in
      List.iter print (make_prod_list 0 "result" pl);
  | STRUCT_TYPE tl ->
      let print (i, n, t) =
	fprintf oc "    %s = " n;
	convert_ml_to_c oc modname (sprintf "Field(v, %d)" i) dir t;
	fprintf oc ";\n"
      in
      List.iter print (make_struct_list 0 "result" tl);
  | CONSTR_TYPE cl ->
      let rec print_cconsts i c = function
	| (CCONST n) :: tl ->
	    fprintf oc "    case %d:\n" c;
	    fprintf oc "        result.tag = %s_%s;\n" modname n;
	    print_cconsts (i + 1) (c + 1) tl
	| (NCCONST(_, _)) :: tl ->
	    print_cconsts (i + 1) c tl
	| [] ->
	    ()
      and print_ncconsts i nc = function
	| (CCONST _) :: tl ->
	    print_ncconsts (i + 1) nc tl
	| (NCCONST(n, PRODUCT_TYPE pl)) :: tl ->
	    let print (i, name, t) =
	      fprintf oc "        %s = " name;
	      convert_ml_to_c oc modname (sprintf "Field(v, %d)" i) dir t;
	      fprintf oc ";\n"
	    in
	    fprintf oc "    case %d:\n" nc;
	    fprintf oc "        result.tag = %s_%s;\n" modname n;
	    let n = sprintf "result.u.%s" n in
	    List.iter print (make_prod_list 0 n pl);
	    print_ncconsts (i + 1) (nc + 1) tl
	| (NCCONST(n, t)) :: tl ->
	    fprintf oc "    case %d:\n" nc;
	    fprintf oc "        result.tag = %s_%s;\n" modname n;
	    fprintf oc "	result.u.%s = " n;
	    convert_ml_to_c oc modname (sprintf "Field(v, %d)" nc) dir t;
	    fprintf oc ";\n";
	    print_ncconsts (i + 1) (nc + 1) tl
	| [] ->
	    ()
      in
      fprintf oc "    if (Is_long(v)) switch (Int_val(v)) {\n";
      print_cconsts 0 0 cl;
      fprintf oc "    default:\n";
      fprintf oc "        failwith(\"bad constant\");\n";
      fprintf oc "    }\n    else switch (Tag_val(v)) {\n";
      print_ncconsts 0 0 cl;
      fprintf oc "    default:\n";
      fprintf oc "        failwith(\"bad tag\");\n    }\n"
  | _ as t ->
      fprintf oc "    result = ";
      convert_ml_to_c oc modname "v" dir t;
      fprintf oc ";\n"

(* This produces a C function that converts an ML value
 * to the corresponding C value.
 *)
let make_ml_to_c_converter env name t =
  let name =
    if name = "t" then
      sprintf "%s_t" env.modname
    else sprintf "%s_%s_t" env.modname name
  in
  fprintf env.doth "%s ml2c_%s(enum camou_direction, value);\n" name name;
  fprintf env.dotc "%s ml2c_%s(enum camou_direction dir, value v){\n    %s result;\n\n"
	name name name;
  ml_to_c env.dotc env.modname name "dir" t;
  fprintf env.dotc "    return result;\n}\n"

(* The list pl is a list of (index, name, type) tuples as created by
 * make_prod_list, make_arg_list, or make_struct_list.  Each entry is
 * a C variable of that type.  Convert those into ML values, and assign
 * them to consecutive fields in the given ML value var.
 *)
let convert_c_to_ml env var indent conv pl =
  let len = List.length pl in
  let fill (i, n, t) =
    let var = if len = 1 then var else (sprintf "Field(%s, %d)" var i) in
    match t with
      | NAMED_TYPE(("", "unit"), []) ->
	  fprintf env.dotc "%s%s = Val_unit;\n" indent var
      | NAMED_TYPE((m, name), []) ->
	  fprintf env.dotc "%s%s = c2ml_" indent var;
	  print_c_name env.dotc env.modname m name;
	  fprintf env.dotc "(%s);\n" n
      | ALPHA_TYPE _ ->
	  fprintf env.dotc "%s%s = c2ml_alpha(%s);\n" indent var n
      | _ as t ->
	  fprintf env.dotc "%s%s = UNKNOWN2(n = %s, t ="
		indent var n;
	  print_ml_type env.dotc env.modname false t;
	  fprintf env.dotc ");\n"
  in
  List.iter fill pl

(* This function takes a type, and produces the body of a function
 * that converts a C value of that type to a corresponding ML value.
 * The signature of that function is void(type v, value *result);
 *)
let c_to_ml env name result callbacks = function
  | NAMED_TYPE((m, n), []) ->
      fprintf env.dotc "    %s = c2ml_" result;
      print_c_name env.dotc env.modname m n;
      fprintf env.dotc "(v);\n"
  | NAMED_TYPE(("", "list"), [t]) ->
      fprintf env.dotc "    if (!camou_ishandle(v, &%s)) {\n" result;
      fprintf env.dotc "        if (v == 0)\n";
      fprintf env.dotc "            %s = Val_int(0);\n" result;
      fprintf env.dotc "        else {\n";
      fprintf env.dotc "            %s = alloc_tuple(2);\n" result;
      convert_c_to_ml env "Field(r[0], 0)" "            " true [0, "v->ent", t];
      fprintf env.dotc "            Field(%s, 1) = c2ml_%s(v->next);\n" result name;
      fprintf env.dotc "        }\n";
      fprintf env.dotc "    }\n"
  | NAMED_TYPE(("", "array"), [t]) ->
      fprintf env.dotc "    if (!camou_ishandle(v, &%s)) {\n" result;
      fprintf env.dotc "        int i;\n\n";
      fprintf env.dotc "        %s = alloc_tuple(v->size);\n" result;
      fprintf env.dotc "        for (i = 0; i < v->size; i++)\n";
      convert_c_to_ml env "Field(r[0], i)" "            " true [0, "v->ent[i]", t];
      fprintf env.dotc "    }\n"
  | CONSTR_TYPE cl ->
      let rec conv_constr c nc = function
	| CCONST name :: tl ->
	    fprintf env.dotc "    case %s_%s:\n" env.modname name;
	    fprintf env.dotc "        r[0] = Val_int(%d);\n" c;
	    fprintf env.dotc "        break;\n";
	    conv_constr (c + 1) nc tl
	| NCCONST(name, PRODUCT_TYPE pl) :: tl ->
	    fprintf env.dotc "    case %s_%s:\n" env.modname name;
	    fprintf env.dotc "        r[0] = alloc(%d, %d);\n" (List.length pl) nc;
	    let name = sprintf "v.u.%s" name in
	    convert_c_to_ml env result "        " true (make_prod_list 0 name pl);
	    fprintf env.dotc "        break;\n";
	    conv_constr c (nc + 1) tl
	| NCCONST(name, NAMED_TYPE((m, t), [])) :: tl ->
	    fprintf env.dotc "    case %s_%s:\n" env.modname name;
	    fprintf env.dotc "        r[0] = alloc(1, %d);\n" nc;
	    fprintf env.dotc "        Field(r[0], 0) = c2ml_";
	    print_c_name env.dotc env.modname m t;
	    fprintf env.dotc "(v.u.%s);\n" name;
	    fprintf env.dotc "        break;\n";
	    conv_constr c (nc + 1) tl
	| NCCONST(name, ALPHA_TYPE _) :: tl ->
	    fprintf env.dotc "    case %s_%s:\n" env.modname name;
	    fprintf env.dotc "        r[0] = alloc(1, %d);\n" nc;
	    fprintf env.dotc "        Field(r[0], 0) = c2ml_alpha(v.u.%s);\n" name;
	    fprintf env.dotc "        break;\n";
	    conv_constr c (nc + 1) tl
	| NCCONST(name, _) :: tl ->
	    fprintf env.dotc "    case %s_%s:\n" env.modname name;
	    fprintf env.dotc "        r[0] = alloc(1, %d);\n" nc;
	    fprintf env.dotc "        ???\n";
	    fprintf env.dotc "        break;\n";
	    conv_constr c (nc + 1) tl
	| [] ->
	    ()
      in
      fprintf env.dotc "    switch (v.tag) {\n";
      conv_constr 0 0 cl;
      fprintf env.dotc "    default:\n";
      fprintf env.dotc "        failwith(\"bad constructor\");\n";
      fprintf env.dotc "    }\n"
  | PRODUCT_TYPE pl ->
      fprintf env.dotc "    r[0] = alloc_tuple(%d);\n" (List.length pl);
      convert_c_to_ml env "r[0]" "    " true (make_prod_list 0 "v" pl);
  | STRUCT_TYPE tl ->
      fprintf env.dotc "    r[0] = alloc_tuple(%d);\n" (List.length tl);
      convert_c_to_ml env "r[0]" "    " true (make_struct_list 0 "v" tl);
  | FUNCTION_TYPE(_, _) ->
      begin
	try
	  let cb = List.assoc name callbacks in
          fprintf env.dotc "    %s = c2ml_callback(cb_%s, v);\n" result cb
	with Not_found ->
	  fprintf env.dotc "    %s = c2ml_callback(cb_<<%s>>, v);\n" result name
      end
  | ALPHA_TYPE _ ->
      fprintf env.dotc "    %s = c2ml_alpha(v);\n" result
  | ABSTRACT_TYPE ->
      fprintf env.dotc "    %s = c2ml_abstract(v);\n" result
  | _ as t ->
      fprintf env.dotc "    %s = UNKNOWN3(t = " result;
      print_ml_type env.dotc env.modname false t;
      fprintf env.dotc ");\n"

let make_c_body env name indent args result =
  let len = List.length args in
  fprintf env.dotc "{\n";
  if result <> unit_type then
    fprintf env.dotc "%s    value result;\n" indent;
  fprintf env.dotc "%s    Push_roots(r, 1);\n\n" indent;
  if len > 1 then
    fprintf env.dotc "%s    r[0] = alloc_tuple(%d);\n" indent len;
  convert_c_to_ml env "r[0]" (indent ^ "    ") false args;
  fprintf env.dotc "%s    " indent;
  if result <> unit_type then
    fprintf env.dotc "result = ";
  fprintf env.dotc "callback(%s, r[0]);\n" name;
  fprintf env.dotc "%s    Pop_roots();\n" indent;
  if result <> unit_type then
    begin
      fprintf env.dotc "%s    return " indent;
      convert_ml_to_c env.dotc env.modname "result" "CAMOU_RESULT" result;
      fprintf env.dotc ";\n"
    end;
  fprintf env.dotc "%s}\n" indent

(* This function produces a C function that saves an ML closure
 * in a global variable, and an ML function that invokes that
 * procedure.
 *)
let register_callback env name t =
  let n = sprintf "%s_%s" env.modname name in
  fprintf env.dotc "static value cb_%s;\n" n;
  fprintf env.dotc "value register_%s(value mlc){\n" n;
  fprintf env.dotc "    register_global_root(&cb_%s);\n" n;
  fprintf env.dotc "    cb_%s = mlc;\n" n;
  fprintf env.dotc "    return Val_unit;\n";
  fprintf env.dotc "}\n";
  fprintf env.dota "external register_%s: " name;
  print_ml_type env.dota env.modname false (FUNCTION_TYPE([t], unit_type));
  fprintf env.dota "\n  = \"register_%s\"\n" n;
  fprintf env.dota "let _ = register_%s %s\n" name name;
  fprintf env.dota "\nexternal addr_%s : unit -> mlc = \"addr_%s\"\n" name n

(* Make a type declaration of the given name and type, and
 * produce conversion functions.  ft is the flattened type, and
 * t the original fully specified type.
 *)
let rec type_decl env name ft t =
  (* This function produces a function that converts a C value to
   * a corresponding ML value.
   *)
  let make_c_to_ml_converter() =
    let n =
      if name = "t" then
        sprintf "%s_t" env.modname
      else sprintf "%s_%s_t" env.modname name
    in
    fprintf env.dotc "/* %s: " n;
    print_ml_type env.dotc env.modname false t;
    fprintf env.dotc " */\n";
    let callbacks =
      match t with
        | FUNCTION_TYPE(args, result) ->
	    make_callback_stub env name args result;
	    [n, sprintf "%s_mlc_%s" env.modname name]
        | _ ->
	    []
    in
    fprintf env.doth "value c2ml_%s(%s v);\n" n n;
    fprintf env.dotc "value c2ml_%s(%s v){\n" n n;
    fprintf env.dotc "    value result;\n";
    fprintf env.dotc "    Push_roots(r, 1);\n\n";
    c_to_ml env n "r[0]" callbacks ft;
    fprintf env.dotc "    result = r[0];\n";
    fprintf env.dotc "    Pop_roots();\n";
    fprintf env.dotc "    return result;\n";
    fprintf env.dotc "}\n"
  in
  local_def := name :: !local_def;
  fprintf env.doth "\ntypedef ";
  let n =
    if name = "t" then
      sprintf "%s_t" env.modname
    else sprintf "%s_%s_t" env.modname name
  in
  print_c_decl env.doth env.modname n "" true ft;
  fprintf env.doth ";\n";
  type_cache := (ft, name) :: !type_cache;
  make_c_to_ml_converter();
  make_ml_to_c_converter env name ft

(* For functions that have callback arguments, this function
 * produces the code that invokes that callback.
 *)
and make_callback_stub env name args result =
  let rec print_arg (_, n, _) =
    fprintf env.dotc "value %s" n
  in
  let stub = sprintf "mlc_%s" name in
  let cbt = FUNCTION_TYPE(NAMED_TYPE(("", "mlc"), []) :: args, result) in
  let flatten = flatten env in
  let args = List.map flatten args in
  let result = flatten result in
  let arg_list = make_arg_list 0 0 args in
  fprintf env.dotc "value %s_%s(value mlc, " env.modname stub;
  print_list env.dotc print_arg ", " arg_list;
  fprintf env.dotc "){\n";
  fprintf env.dotc "    %s_%s_t upcall = (%s_%s_t) mlc;\n"
	env.modname name env.modname name;
  if result = unit_type then
    fprintf env.dotc "\n    "
  else
    begin
      print_c_decl env.dotc env.modname "v" "    " true result;
      fprintf env.dotc " = "
    end;
  fprintf env.dotc "(*upcall)(";
  let convert (_, n, t) =
    convert_ml_to_c env.dotc env.modname n "CAMOU_ARG" t
  in
  print_list env.dotc convert ", " arg_list;
  fprintf env.dotc ");\n";
  if result = unit_type then
    fprintf env.dotc "    return Val_unit;\n"
  else
    begin
      fprintf env.dotc "    value result;\n\n";
      c_to_ml env stub "result" [] result;
      fprintf env.dotc "    return result;\n"
    end;
  fprintf env.dotc "}\n";
  let n = sprintf "%s_callback_%s" env.modname name in
  let t = FUNCTION_TYPE(FUNCTION_TYPE(args, result) :: args, result) in
  print_c_decl env.dotc env.modname n "" false t;
  fprintf env.dotc "{\n    value v;\n";
  fprintf env.dotc "    if (camou_ishandle(a0, &v)) ";
  make_c_body env "v" "    " (make_arg_list 0 1 args) result;
  fprintf env.dotc "    else\n";
  fprintf env.dotc "        ";
  if result <> unit_type then
    fprintf env.dotc "return ";
  fprintf env.dotc "(*a0)(";
  if args <> [unit_type] then
    begin
      let print_arg (i, _, _) = fprintf env.dotc "a%d" (i + 1) in
      print_list env.dotc print_arg ", " arg_list
    end;
  fprintf env.dotc ");\n";
  fprintf env.dotc "}\n";
  fprintf env.dota "\nexternal %s: " stub;
  print_ml_type env.dota env.modname false cbt;
  fprintf env.dota "\n  = \"%s_%s\"\n" env.modname stub;
  register_callback env stub cbt

and flatten env t =
  let make_stub ft t =
    let rec find p = function
      | (x, n) :: tl ->
        if p x then
	  n
	else
	  find p tl
      | [] ->
	  raise Not_found
    in
    let pred t = type_cmp (t, ft) in
    let stub =
      try
	find pred !type_cache
      with Not_found ->
        let stub = sprintf "uid_%d" !uid_counter in
        incr uid_counter;
        type_decl env stub ft t;
	stub
    in
    NAMED_TYPE(("", stub), [])
  in
  let rec doit = function
    | NAMED_TYPE(n, []) ->
        NAMED_TYPE(n, [])
    | NAMED_TYPE(n, nl) as t ->
	let nl = List.map doit nl in
	make_stub (NAMED_TYPE(n, nl)) t
    | FUNCTION_TYPE(args, result) as t ->
	let args = List.map doit args in
        make_stub (FUNCTION_TYPE(args, doit result)) t
    | PRODUCT_TYPE pl as t->
	let pl = List.map doit pl in
	make_stub (PRODUCT_TYPE pl) t
    | STRUCT_TYPE sl as t ->
	let flatten (n, m, t) = (n, m, doit t) in
	make_stub (STRUCT_TYPE (List.map flatten sl)) t
    | CONSTR_TYPE cl as t ->
        let flatten = function
          | CCONST _ as c ->
	      c
	  | NCCONST(n, PRODUCT_TYPE tl) ->
	      let tl = List.map doit tl in
	      NCCONST(n, PRODUCT_TYPE tl)
	  | NCCONST(n, t) ->
	      NCCONST(n, doit t)
	in
	let cl = List.map flatten cl in
	make_stub (CONSTR_TYPE cl) t
    | _ as t ->
	t
  in doit t

let flatten2 env params = function
  | NAMED_TYPE(n, nl) ->
      let nl = List.map (flatten env) nl in
      NAMED_TYPE(n, nl)
  | FUNCTION_TYPE(args, result) as t ->
      let args = List.map (flatten env) args in
      let result = flatten env result in
      FUNCTION_TYPE(args, result)
  | PRODUCT_TYPE tl ->
      let tl = List.map (flatten env) tl in
      PRODUCT_TYPE tl
  | STRUCT_TYPE sl ->
      let flatten (n, m, t) =
	(n, m, flatten env t)
      in
      STRUCT_TYPE (List.map flatten sl)
    | CONSTR_TYPE cl ->
        let flatten = function
          | CCONST _ as c ->
	      c
	  | NCCONST(n, PRODUCT_TYPE tl) ->
	      let tl =
		List.map (flatten env) tl
	      in
	      NCCONST(n, PRODUCT_TYPE tl)
	  | NCCONST(n, t) ->
	      NCCONST(n, flatten env t)
	in
	CONSTR_TYPE (List.map flatten cl)
  | _ as t ->
      t

(* This function produces a C interface to an ML function
 * of the given type t.
 *)
let make_c_to_ml_stub env name t =
  let check_func = function
    | FUNCTION_TYPE(args, result) as t ->
	let cbt =
          if List.length args > 1 then
            FUNCTION_TYPE([PRODUCT_TYPE args], result)
	  else
	    t
	in
	let arg_list = make_arg_list 0 0 args in
	fprintf env.dota "\nlet %s (" name;
	let print_arg (_, n, _) = fprintf env.dota "%s" n in
	print_list env.dota print_arg ", " arg_list;
	fprintf env.dota ") = %s.%s" env.modname name;
	let print_arg (_, n, _) = fprintf env.dota " %s" n in
	List.iter print_arg arg_list;
	fprintf env.dota "\n";
	register_callback env name cbt;
	fprintf env.dota "let _ = %s_pre.wrap " env.modname;
	fprintf env.dota "%s.%s (addr_%s())\n" env.modname name name;
	let flatten = flatten env in
	let args = List.map flatten args in
	let result = flatten result in
	let t = FUNCTION_TYPE(args, result) in
	let args = make_arg_list 0 0 args in
	let name = sprintf "%s_%s" env.modname name in
	fprintf env.doth "\n";
	print_c_decl env.doth env.modname name "" false t;
	fprintf env.doth ";\n";
        print_c_decl env.dotc env.modname name "" false t;
        make_c_body env (sprintf "cb_%s" name) "" args result;
	fprintf env.dotc "value addr_%s(value unit){\n" name;
	fprintf env.dotc "    value closure = alloc(1, Closure_tag + 1);\n";
	fprintf env.dotc "    Field(closure, 0) = (value) %s;\n" name;
	fprintf env.dotc "    return closure;\n";
	fprintf env.dotc "}\n"
    | _ ->
        fprintf env.dotc "/* %s: can only make stubs for functions */\n" name
  in
  local_def := name :: !local_def;
  check_func t

(* This creates a C function that may be invoked from ML.
 *)
let make_ml_to_c_stub env name extname =
  let make_c_func args result =
    let rec print_arg (_, n, _) =
      fprintf env.dotc "value %s" n
    in
    let cbargs = args and cbresult = result in
    let flatten = flatten env in
    let args = List.map flatten args in
    let result = flatten result in
    let arg_list = make_arg_list 0 0 args in
    fprintf env.dotc "value %s_%s(" env.modname name;
    print_list env.dotc print_arg ", " arg_list;
    fprintf env.dotc "){\n";
    if result = unit_type then
      fprintf env.dotc "    "
    else
      begin
        print_c_decl env.dotc env.modname "v" "    " true result;
        fprintf env.dotc " = "
      end;
    fprintf env.dotc "%s(" extname;
    let convert (_, n, t) =
      convert_ml_to_c env.dotc env.modname n "CAMOU_ARG" t
    in
    print_list env.dotc convert ", " arg_list;
    fprintf env.dotc ");\n";
    if result = unit_type then
      fprintf env.dotc "    return Val_unit;\n"
    else
      begin
        fprintf env.dotc "    value result;\n\n";
        c_to_ml env name "result" [] result;
        fprintf env.dotc "    return result;\n"
      end;
    fprintf env.dotc "}\n";
    fprintf env.dotb "\nexternal %s: " name;
    let print_arg t =
      print_ml_type env.dotb env.modname true t;
      fprintf env.dotb " -> "
    in
    List.iter print_arg cbargs;
    print_ml_type env.dotb env.modname true cbresult;
    fprintf env.dotb "\n  = \"%s_%s\"\n" env.modname name
  in function
  | FUNCTION_TYPE(args, result) ->
      make_c_func args result
  | _ ->
      failwith "can only make stubs for functions"

let check_includes env =
  let rec check = function
    | FUNCTION_TYPE(args, result) ->
        List.iter check args;
        check result
    | PRODUCT_TYPE pl ->
        List.iter check pl;
    | NAMED_TYPE((m, n), tl) ->
        List.iter check tl;
        if (m <> "") && not (List.mem m !included) then
	  begin
            fprintf env.dotc "\n#include \"%s.h\""
		(uncapitalize m);
	    included := m :: !included
	  end
    | CONSTR_TYPE cl ->
        let check = function
          | CCONST _ -> ()
          | NCCONST(_, t) -> check t
        in List.iter check cl
    | STRUCT_TYPE sl ->
        let check (_, _, t) = check t in
        List.iter check sl
    | _ ->
        ()
  in function
  | TYPE_DECL(_, t) ->
      check t
  | VAL_DECL(_, t) ->
      check t
  | INCLUDE_DECL("h", what) ->
      fprintf env.doth "%s" what
  | INCLUDE_DECL("c", what) ->
      fprintf env.dotc "%s" what
  | INCLUDE_DECL("pre", what) ->
      fprintf env.dotb "%s" what
  | INCLUDE_DECL("post", what) ->
      fprintf env.dota "%s" what
  | _ ->
      ()

(* This function takes a type or function declaration from the
 * input .mli file, and deals with it.
 *)
let handle env = function
  | TYPE_DECL((name, params), t) ->
      fprintf env.dotc "\n/* type %s */\n" name;
      let ft = flatten2 env params t in
      type_decl env name ft t
  | VAL_DECL(name, t) ->
      fprintf env.dotc "\n/* val %s */\n" name;
      make_c_to_ml_stub env name t
  | EXTTYPE_DECL(name, extname, t) ->
      fprintf env.dotc "\n/* exttype %s */\n" name;
      if t = ABSTRACT_TYPE then
        begin
          fprintf env.doth "typedef %s %s;\n" extname name;
          fprintf env.doth "#define c2ml_%s(v) ((%s) (v))\n" name extname;
          fprintf env.doth "#define ml2c_%s(v) ((value) (v))\n" name;
          fprintf env.dota "\ntype %s\n" name
        end
      else
	begin
          let ft = flatten2 env [] t in
          type_decl env name ft t;
	  fprintf env.dota "\ntype %s = " name;
	  print_ml_type env.dota env.modname false t;
	  fprintf env.dota "\n"
	end
  | EXTVAL_DECL(name, extname, t) ->
      fprintf env.dotc "\n/* external %s */\n" name;
      make_ml_to_c_stub env name extname t
  | EXCEPTION_DECL _ ->
      fprintf env.dotc "\n/* Exception declarations are ignored */\n"
  | _ ->
      ()

(* Put headers on some files.
 *)
let init env =
  fprintf env.dotc "#include \"memory.h\"\n";
  fprintf env.dotc "#include \"camou.h\"\n";
  fprintf env.dotb "type mlc\n\n";
  fprintf env.dotb "let wrap_table = ref ([] : (Obj.t * mlc) list)\n";
  fprintf env.dotb "let wrap f a = wrap_table := (Obj.repr f, a) :: !wrap_table\n";
  fprintf env.dotb "let magic o = List.assq (Obj.repr o) !wrap_table\n";
  fprintf env.dota "type mlc = %s_pre.mlc\n" env.modname

(* Parse the given input file and produce code from it.
 *)
let process file =
  uid_counter := 0;
  local_def := [];
  included := [];
  type_cache := [];
  if not (Filename.check_suffix file ".mli") then
    failwith "file doesn't have suffix .mli";
  let ic = open_in file in
  let lexbuf = Lexing.from_channel ic in
  let decl = try
    Mlipars.interface Mlilex.token lexbuf
  with Parsing.Parse_error ->
    printf "parse error in line %d\n" !Mlilex.lineno;
    raise Parsing.Parse_error
  in
  close_in ic;
  let file = Filename.basename file in
  let umodname = Filename.chop_suffix file ".mli" in
  let doth = open_out (umodname ^ ".h")
  and dotc = open_out (umodname ^ ".c")
  and dotb = open_out (umodname ^ "_pre.ml")
  and dota = open_out (umodname ^ "_post.ml")
  in
  let modname = capitalize umodname in
  let env = {
    dota = dota; dotb = dotb; dotc = dotc; doth = doth;
    modname = modname
  } in
  init env;
  List.iter (check_includes env) decl;
  fprintf dotc "\n#include \"%s.h\"\n" umodname;
  List.iter (handle env) decl;
  close_out doth;
  close_out dotc;
  close_out dotb;
  close_out dota

let main _ =
  Arg.parse [] process "typ"

let _ = Printexc.catch main()
